import time
import logging
import contextlib
import pg_execute_values as pev

logging.basicConfig(level=logging.INFO)
log = logging.getLogger("metrics")

def report_metrics(op, sql, op_time):
    log.info(f"Operation: {op.name!r}\nSQL: {sql!r} \nTime (ms): {op_time}")

@contextlib.contextmanager
def observe_query(func):
    op = func.operation
    sql = func.sql
    start = time.time()
    yield
    end = time.time()
    op_time = end - start
    report_metrics(op, sql, op_time)

with observe_query(pev.queries.getem):
    pev.queries.getem(pev.conn)

# INFO:metrics:Operation: 'SELECT'
# SQL: 'select * from test order by id;' 
# Time (ms): 2.6226043701171875e-06



from .generic import GenericAdapter
from ..utils import VAR_REF
from typing import List


def _colon_to_dollar(ma):
    """Convert 'WHERE :id = 1' to 'WHERE $id = 1'."""
    gd = ma.groupdict()
    if gd["dquote"] is not None:
        return gd["dquote"]
    elif gd["squote"] is not None:
        return gd["squote"]
    else:
        return f'{gd["lead"]}${gd["var_name"]}'


class DuckDBAdapter(GenericAdapter):
    """DuckDB Adapter"""

    def __init__(self, driver=None, cursor_as_dict: bool = False):
        super().__init__(driver=driver)
        # whether to converts the default tuple response to a dict.
        self._convert_row_to_dict = cursor_as_dict

    def process_sql(self, _query_name, _op_type, sql):
        return VAR_REF.sub(_colon_to_dollar, sql)

    def insert_returning(self, conn, _query_name, sql, parameters):  # pragma: no cover
        # very similar to select_one but the returned value
        cur = self._cursor(conn)
        cur.execute(sql, parameters)
        # we have to use fetchall instead of fetchone for now due to this:
        # https://github.com/duckdb/duckdb/issues/6008
        res = cur.fetchall()
        cur.close()
        if isinstance(res, list):
            res = res[0]
        return res[0] if res and len(res) == 1 else res

    def select(self, conn, _query_name: str, sql: str, parameters, record_class=None):
        column_names: List[str] = []
        cur = self._cursor(conn)
        try:
            cur.execute(sql, parameters)
            if record_class is None:
                first = True
                for row in cur.fetchall():
                    if first:  # get column names on the fly
                        column_names = [c[0] for c in cur.description or []]
                        first = False
                    if self._convert_row_to_dict:  # pragma: no cover
                        # strict=False: requires 3.10
                        yield dict(zip(column_names, row))
                    else:
                        yield row
            else:  # pragma: no cover
                first = True
                for row in cur.fetchall():
                    if first:  # only get description on the fly, for apsw
                        column_names = [c[0] for c in cur.description or []]
                        first = False
                    # strict=False: requires 3.10
                    yield record_class(**dict(zip(column_names, row)))
        finally:
            cur.close()

    def select_one(self, conn, _query_name, sql, parameters, record_class=None):
        cur = self._cursor(conn)
        try:
            cur.execute(sql, parameters)
            result = cur.fetchone()
            if result is not None and record_class is not None:  # pragma: no cover
                column_names = [c[0] for c in cur.description or []]
                # strict=False: requires 3.10
                result = record_class(**dict(zip(column_names, result)))
            elif result is not None and self._convert_row_to_dict:  # pragma: no cover
                column_names = [c[0] for c in cur.description or []]
                result = dict(zip(column_names, result))
        finally:
            cur.close()
        return result



import re
import logging

# FIXME to be improved
VAR_REF = re.compile(
    # NOTE probably pg specific?
    r'(?P<dquote>"(""|[^"])+")|'
    # FIXME mysql/mariadb use backslash escapes
    r"(?P<squote>\'(\'\'|[^\'])*\')|"
    # NOTE beware of overlapping re
    r"(?P<lead>[^:]):(?P<var_name>\w+)(?=[^:]?)"
)
"""Pattern to identify colon-variables (aka _named_ style) in SQL code"""

# NOTE see comments above
VAR_REF_DOT = re.compile(
    r'(?P<dquote>"(""|[^"])+")|'
    r"(?P<squote>\'(\'\'|[^\'])*\')|"
    r"(?P<lead>[^:]):(?P<var_name>\w+\.\w+)(?=[^:]?)"
)
"""Pattern to identify colon-variables with a simple attribute in SQL code."""

log = logging.getLogger("aiosql")
"""Shared package logging."""
# log.setLevel(logging.DEBUG)


class SQLLoadException(Exception):
    """Raised when there is a problem loading SQL content from a file or directory"""

    pass


class SQLParseException(Exception):
    """Raised when there was a problem parsing the aiosql comment annotations in SQL"""

    pass



from datetime import date

import aiosql
import pytest
import run_tests as t
import utils as u

try:
    import psycopg as db
    from psycopg.rows import dict_row
except ModuleNotFoundError:
    pytest.skip("missing driver: psycopg", allow_module_level=True)

pytestmark = [
    pytest.mark.postgres,
    pytest.mark.skipif(not u.has_pkg("pytest_postgresql"), reason="no pytest_postgresql"),
]

DRIVER = "psycopg"


@pytest.fixture
def queries():
    return t.queries(DRIVER)


@pytest.fixture
def conn(pg_conn):
    return pg_conn


def test_version():
    assert db.__version__.startswith("3.")


def test_cursor(conn, queries):
    t.run_cursor(conn, queries)


def test_record_query(pg_params, queries):
    with db.connect(**pg_params, row_factory=dict_row) as conn:
        t.run_record_query(conn, queries)
        # test select_value with dict_row
        t.run_select_value(conn, queries, DRIVER)


def test_parameterized_query(conn, queries):
    t.run_parameterized_query(conn, queries)


def test_parameterized_record_query(pg_params, queries):
    with db.connect(**pg_params, row_factory=dict_row) as conn:
        t.run_parameterized_record_query(conn, queries, DRIVER, date)


def test_record_class_query(conn, queries):
    t.run_record_class_query(conn, queries, date)


def test_select_cursor_context_manager(conn, queries):
    t.run_select_cursor_context_manager(conn, queries, date)


def test_select_one(conn, queries):
    t.run_select_one(conn, queries)


def test_select_value(conn, queries):
    t.run_select_value(conn, queries, DRIVER)


def test_modulo(conn, queries):
    actual = queries.blogs.pg_get_modulo(conn, numerator=7, denominator=3)
    expected = 7 % 3
    assert actual == expected


def test_insert_returning(conn, queries):
    t.run_insert_returning(conn, queries, DRIVER, date)


def test_delete(conn, queries):
    t.run_delete(conn, queries)


def test_insert_many(conn, queries):
    t.run_insert_many(conn, queries, date)


def test_date_time(conn, queries):
    t.run_date_time(conn, queries, DRIVER)


def test_execute_script(conn, queries):
    actual = queries.comments.pg_create_comments_table(conn)
    assert actual == "CREATE TABLE"


def test_object_attributes(conn, queries):
    t.run_object_attributes(conn, queries, DRIVER)



import aiosql
import pytest
import run_tests as t
import utils as u

try:
    import aiosqlite
except ModuleNotFoundError:
    pytest.skip("missing driver: aiosqlite", allow_module_level=True)

pytestmark = [
    pytest.mark.sqlite3,
    # pytest.mark.asyncio,
    pytest.mark.skipif(not u.has_pkg("pytest_asyncio"), reason="no pytest_asyncio"),
]

DRIVER = "aiosqlite"


@pytest.fixture
def queries():
    return t.queries("aiosqlite")


def dict_factory(cursor, row):
    d = {}
    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d


# NOTE no cursor on aiosqlite driver adapter
# def test_cursor(conn, queries):
#    t.run_cursor(conn, queries)


@pytest.mark.asyncio
async def test_record_query(sqlite3_db_path, queries):
    async with aiosqlite.connect(sqlite3_db_path) as conn:
        conn.row_factory = dict_factory
        await t.run_async_record_query(conn, queries)


@pytest.mark.asyncio
async def test_parameterized_query(sqlite3_db_path, queries):
    async with aiosqlite.connect(sqlite3_db_path) as conn:
        await t.run_async_parameterized_query(conn, queries, t.todate)


@pytest.mark.asyncio
async def test_parameterized_record_query(sqlite3_db_path, queries):
    async with aiosqlite.connect(sqlite3_db_path) as conn:
        conn.row_factory = dict_factory
        await t.run_async_parameterized_record_query(conn, queries, DRIVER, t.todate)


@pytest.mark.asyncio
async def test_record_class_query(sqlite3_db_path, queries):
    async with aiosqlite.connect(sqlite3_db_path) as conn:
        await t.run_async_record_class_query(conn, queries, t.todate)


@pytest.mark.asyncio
async def test_select_cursor_context_manager(sqlite3_db_path, queries):
    async with aiosqlite.connect(sqlite3_db_path) as conn:
        await t.run_async_select_cursor_context_manager(conn, queries, t.todate)


@pytest.mark.asyncio
async def test_select_one(sqlite3_db_path, queries):
    async with aiosqlite.connect(sqlite3_db_path) as conn:
        await t.run_async_select_one(conn, queries)


@pytest.mark.asyncio
async def test_select_value(sqlite3_db_path, queries):
    async with aiosqlite.connect(sqlite3_db_path) as conn:
        await t.run_async_select_value(conn, queries)


@pytest.mark.asyncio
async def test_insert_returning(sqlite3_db_path, queries):
    async with aiosqlite.connect(sqlite3_db_path) as conn:
        await t.run_async_insert_returning(conn, queries, DRIVER, t.todate)


@pytest.mark.asyncio
async def test_delete(sqlite3_db_path, queries):
    async with aiosqlite.connect(sqlite3_db_path) as conn:
        await t.run_async_delete(conn, queries)


@pytest.mark.asyncio
async def test_insert_many(sqlite3_db_path, queries):
    async with aiosqlite.connect(sqlite3_db_path) as conn:
        await t.run_async_insert_many(conn, queries, t.todate)


@pytest.mark.asyncio
async def test_async_methods(sqlite3_db_path, queries):
    async with aiosqlite.connect(sqlite3_db_path) as conn:
        conn.row_factory = dict_factory
        await t.run_async_methods(conn, queries)


@pytest.mark.asyncio
async def test_execute_script(sqlite3_db_path, queries):
    async with aiosqlite.connect(sqlite3_db_path) as conn:
        actual = await queries.comments.sqlite_create_comments_table(conn)
        assert actual == "DONE"



import pytest
from pathlib import Path
from conf_schema import create_user_blogs, USERS_DATA_PATH, BLOGS_DATA_PATH

try:
    import duckdb

    @pytest.fixture
    def duckdb_db_path(tmpdir):
        db_path = str(Path(tmpdir.strpath) / "blogdb.duck.db")
        return db_path

    @pytest.fixture
    def duckdb_conn(duckdb_db_path):
        conn = duckdb.connect(":memory:")
        populate_duckdb_db(conn)
        yield conn
        conn.close()

    def populate_duckdb_db(conn):
        conn.execute("\n".join(create_user_blogs("duckdb")))
        conn.execute(
            "insert into users(userid, username, firstname, lastname)\n"
            "select nextval('users_seq') as userid, column0 as username,\n"
            "       column1 as firstname, column2 as lastname\n"
            f"from read_csv_auto('{str(USERS_DATA_PATH)}', header=False)"
        )
        conn.execute(
            "insert into blogs(blogid, userid, title, content, published)\n"
            "select nextval('blogs_seq') as blogid, column0 as userid, column1 as title,\n"
            "       column2 as content, column3 as published\n"
            f"from read_csv_auto('{str(BLOGS_DATA_PATH)}', header=False)"
        )

except ModuleNotFoundError:

    @pytest.fixture
    def duckdb_db_path(tmpdir):
        raise Exception("undefined fixture")

    @pytest.fixture
    def duckdb_conn(duckdb_db_path):
        raise Exception("undefined fixture")



from contextlib import asynccontextmanager


class AioSQLiteAdapter:
    is_aio_driver = True

    def process_sql(self, _query_name, _op_type, sql):
        """Pass through function because the ``aiosqlite`` driver can already handle the
        ``:var_name`` format used by aiosql and doesn't need any additional processing.

        Args:

        - _query_name (str): The name of the sql query.
        - _op_type (SQLOperationType): The type of SQL operation performed by the query.
        - sql (str): The sql as written before processing.

        Returns:

        - str: Original SQL text unchanged.
        """
        return sql

    async def select(self, conn, _query_name, sql, parameters, record_class=None):
        async with conn.execute(sql, parameters) as cur:
            results = await cur.fetchall()
            if record_class is not None:
                column_names = [c[0] for c in cur.description]
                results = [record_class(**dict(zip(column_names, row))) for row in results]
        return results

    async def select_one(self, conn, _query_name, sql, parameters, record_class=None):
        async with conn.execute(sql, parameters) as cur:
            result = await cur.fetchone()
            if result is not None and record_class is not None:
                column_names = [c[0] for c in cur.description]
                result = record_class(**dict(zip(column_names, result)))
        return result

    async def select_value(self, conn, _query_name, sql, parameters):
        async with conn.execute(sql, parameters) as cur:
            result = await cur.fetchone()
        return result[0] if result else None

    @asynccontextmanager
    async def select_cursor(self, conn, _query_name, sql, parameters):
        async with conn.execute(sql, parameters) as cur:
            yield cur

    async def insert_returning(self, conn, _query_name, sql, parameters):
        async with conn.execute(sql, parameters) as cur:
            return cur.lastrowid

    async def insert_update_delete(self, conn, _query_name, sql, parameters):
        async with conn.execute(sql, parameters) as cur:
            return cur.rowcount

    async def insert_update_delete_many(self, conn, _query_name, sql, parameters):
        cur = await conn.executemany(sql, parameters)
        await cur.close()

    async def execute_script(self, conn, sql):
        await conn.executescript(sql)
        return "DONE"



from datetime import date

import aiosql
import pytest
import run_tests as t
import utils as u

try:
    import pymysql as db
except ModuleNotFoundError:
    pytest.skip("missing driver: pymysql", allow_module_level=True)

DRIVER = "pymysql"

pytestmark = [
    pytest.mark.mysql,
    pytest.mark.skipif(not u.has_pkg("pytest_mysql"), reason="no pytest_mysql"),
]


@pytest.fixture
def queries():
    return t.queries(DRIVER)


@pytest.fixture
def conn(my_conn):
    return my_conn


@pytest.fixture
def conn_db(my_db):
    return my_db


def test_my_dsn(my_dsn):
    assert "user" in my_dsn and "host" in my_dsn and "port" in my_dsn


def test_my_conn(conn):
    assert conn.__module__.startswith(db.__name__)
    t.run_something(conn)


def test_my_db(conn_db):
    assert conn_db.__module__.startswith(db.__name__)
    t.run_something(conn_db)


def test_cursor(conn, queries):
    t.run_cursor(conn, queries)


def test_record_query(conn_db, my_dsn, queries):
    with db.connect(**my_dsn, cursorclass=db.cursors.DictCursor) as conn:
        t.run_record_query(conn, queries)


def test_parameterized_query(conn_db, my_dsn, queries):
    t.run_parameterized_query(conn_db, queries)


@pytest.mark.skip("pymysql issue when mogrifying because of date stuff %Y")
def test_parameterized_record_query(conn_db, my_dsn, queries):  # pragma: no cover
    with db.connect(**pymysql_db_dsn, cursorclass=db.cursors.DictCursor) as conn:
        t.run_parameterized_record_query(conn, queries, DRIVER, date)


def test_record_class_query(conn_db, queries):
    t.run_record_class_query(conn_db, queries, date)


def test_select_cursor_context_manager(conn_db, queries):
    t.run_select_cursor_context_manager(conn_db, queries, date)


def test_select_one(conn_db, queries):
    t.run_select_one(conn_db, queries)


def test_select_value(conn_db, queries):
    t.run_select_value(conn_db, queries, DRIVER)


@pytest.mark.skip("MySQL does not support RETURNING")
def test_insert_returning(conn_db, queries):  # pragma: no cover
    t.run_insert_returning(conn_db, queries, DRIVER, date)


def test_delete(conn_db, queries):
    t.run_delete(conn_db, queries)


def test_insert_many(conn_db, queries):
    t.run_insert_many(conn_db, queries, date)


def test_date_time(conn_db, queries):
    t.run_date_time(conn_db, queries, DRIVER)



from .pyformat import PyFormatAdapter


class BrokenMySQLAdapter(PyFormatAdapter):
    """
    Work around PyMySQL and MySQLDB mishandling of empty parameters
    and lack of willingness to fix the issue.

    See: https://github.com/PyMySQL/PyMySQL/issues/1059
    """

    def select(self, conn, name, sql, parameters, record_class=None):
        return super().select(conn, name, sql, parameters or None, record_class)

    def select_one(self, conn, name, sql, parameters, record_class=None):
        return super().select_one(conn, name, sql, parameters or None, record_class)

    def select_value(self, conn, name, sql, parameters):
        return super().select_value(conn, name, sql, parameters or None)

    def insert_update_delete(self, conn, name, sql, parameters):
        return super().insert_update_delete(conn, name, sql, parameters or None)

    # only called for mariadb, as mysql does not implement RETURNING
    def insert_returning(self, conn, name, sql, parameters):  # pragma: no cover
        return super().insert_returning(conn, name, sql, parameters or None)

    # left out with parameters: insert_update_delete_many, select_cursor



import inspect
from pathlib import Path
from types import MethodType

# TODO drop most of this ugly stuff when >= 3.10
from typing import Any, Callable, List, Optional, Set, Tuple, Union, Dict, cast

from .types import DriverAdapterProtocol, QueryDatum, QueryDataTree, QueryFn, SQLOperationType
from .utils import SQLLoadException


class Queries:
    """Container object with dynamic methods built from SQL queries.

    The ``-- name:`` definition comments in the content of the SQL determine what the dynamic
    methods of this class will be named.

    Much of the needed pre-processing is performed in ``QueryLoader``.

    **Parameters:**

    - **driver_adapter**: Either a string to designate one of the aiosql built-in database driver
      adapters (e.g. "sqlite3", "psycopg").
      If you have defined your own adapter class, you can pass its constructor.
    - **kwargs_only**: whether to reject positional parameters, defaults to false.
    """

    def __init__(
        self,
        driver_adapter: DriverAdapterProtocol,
        kwargs_only: bool = False,
    ):
        self.driver_adapter: DriverAdapterProtocol = driver_adapter
        self.is_aio: bool = getattr(driver_adapter, "is_aio_driver", False)
        self._kwargs_only = kwargs_only
        self._available_queries: Set[str] = set()

    #
    # INTERNAL UTILS
    #
    def _params(
        self, attributes, args: Union[List[Any], Tuple[Any]], kwargs: Dict[str, Any]
    ) -> Union[List[Any], Tuple[Any], Dict[str, Any]]:
        """Handle query parameters.

        - update attribute references ``:u.a`` to ``:u__a``.
        - check whether non kwargs are allowed and other checks.
        - return the parameters, either ``args`` or ``kwargs``.
        """

        if attributes and kwargs:

            # switch o.a to o<attribute>a
            for var, atts in attributes.items():
                if var not in kwargs:
                    raise ValueError(f"missing named parameter {var}")
                val = kwargs.pop(var)
                for att, var_name in atts.items():
                    if not hasattr(val, att):
                        raise ValueError(f"parameter {var} is missing attribute {att}")
                    kwargs[var_name] = getattr(val, att)

        if self._kwargs_only:
            if args:
                raise ValueError("cannot use positional parameters under kwargs_only")
            return kwargs
        elif kwargs:
            # FIXME is this true?
            if args:
                raise ValueError("cannot mix positional and named parameters in query")
            return kwargs
        else:
            return args

    def _query_fn(
        self,
        fn: Callable[..., Any],
        name: str,
        doc: Optional[str],
        sql: str,
        operation: SQLOperationType,
        signature: Optional[inspect.Signature],
        floc: Tuple[Union[Path, str], int] = ("<unknown>", 0),
        attributes: Optional[Dict[str, Dict[str, str]]] = None,
    ) -> QueryFn:
        """Add custom-made metadata to a dynamically generated function."""
        fname, lineno = floc
        fn.__code__ = fn.__code__.replace(co_filename=str(fname), co_firstlineno=lineno)  # type: ignore
        qfn = cast(QueryFn, fn)
        qfn.__name__ = name
        qfn.__doc__ = doc
        qfn.__signature__ = signature
        qfn.sql = sql
        qfn.operation = operation
        qfn.attributes = attributes
        return qfn

    # NOTE about coverage: because __code__ is set to reflect the actual SQL file
    # source, coverage does note detect that the "fn" functions are actually called,
    # hence the "no cover" hints.
    def _make_sync_fn(self, query_datum: QueryDatum) -> QueryFn:
        """Build a dynamic method from a parsed query."""

        query_name, doc_comments, operation_type, sql, record_class, signature, floc, attributes = (
            query_datum
        )

        if operation_type == SQLOperationType.INSERT_RETURNING:

            def fn(self, conn, *args, **kwargs):  # pragma: no cover
                return self.driver_adapter.insert_returning(
                    conn, query_name, sql, self._params(attributes, args, kwargs)
                )

        elif operation_type == SQLOperationType.INSERT_UPDATE_DELETE:

            def fn(self, conn, *args, **kwargs):  # type: ignore # pragma: no cover
                return self.driver_adapter.insert_update_delete(
                    conn, query_name, sql, self._params(attributes, args, kwargs)
                )

        elif operation_type == SQLOperationType.INSERT_UPDATE_DELETE_MANY:

            def fn(self, conn, *args, **kwargs):  # type: ignore # pragma: no cover
                assert not kwargs, "cannot use named parameters in many query"  # help type checker
                return self.driver_adapter.insert_update_delete_many(conn, query_name, sql, *args)

        elif operation_type == SQLOperationType.SCRIPT:

            def fn(self, conn, *args, **kwargs):  # type: ignore # pragma: no cover
                # FIXME parameters are ignored?
                return self.driver_adapter.execute_script(conn, sql)

        elif operation_type == SQLOperationType.SELECT:

            def fn(self, conn, *args, **kwargs):  # type: ignore # pragma: no cover
                return self.driver_adapter.select(
                    conn, query_name, sql, self._params(attributes, args, kwargs), record_class
                )

        elif operation_type == SQLOperationType.SELECT_ONE:

            def fn(self, conn, *args, **kwargs):  # pragma: no cover
                return self.driver_adapter.select_one(
                    conn, query_name, sql, self._params(attributes, args, kwargs), record_class
                )

        elif operation_type == SQLOperationType.SELECT_VALUE:

            def fn(self, conn, *args, **kwargs):  # pragma: no cover
                return self.driver_adapter.select_value(
                    conn, query_name, sql, self._params(attributes, args, kwargs)
                )

        else:
            raise ValueError(f"Unknown operation_type: {operation_type}")

        return self._query_fn(
            fn, query_name, doc_comments, sql, operation_type, signature, floc, attributes
        )

    # NOTE does this make sense?
    def _make_async_fn(self, fn: QueryFn) -> QueryFn:
        """Wrap in an async function."""

        async def afn(self, conn, *args, **kwargs):  # pragma: no cover
            return await fn(self, conn, *args, **kwargs)

        return self._query_fn(afn, fn.__name__, fn.__doc__, fn.sql, fn.operation, fn.__signature__)

    def _make_ctx_mgr(self, fn: QueryFn) -> QueryFn:
        """Wrap in a context manager function."""

        def ctx_mgr(self, conn, *args, **kwargs):  # pragma: no cover
            return self.driver_adapter.select_cursor(
                conn, fn.__name__, fn.sql, self._params(fn.attributes, args, kwargs)
            )

        return self._query_fn(
            ctx_mgr, f"{fn.__name__}_cursor", fn.__doc__, fn.sql, fn.operation, fn.__signature__
        )

    def _create_methods(self, query_datum: QueryDatum, is_aio: bool) -> List[QueryFn]:
        """Internal function to feed add_queries."""
        fn = self._make_sync_fn(query_datum)
        if is_aio:
            fn = self._make_async_fn(fn)

        ctx_mgr = self._make_ctx_mgr(fn)

        if query_datum.operation_type == SQLOperationType.SELECT:
            return [fn, ctx_mgr]
        else:
            return [fn]

    #
    # PUBLIC INTERFACE
    #
    @property
    def available_queries(self) -> List[str]:
        """Returns listing of all the available query methods loaded in this class.

        **Returns:** ``list[str]`` List of dot-separated method accessor names.
        """
        return sorted(self._available_queries)

    def __repr__(self) -> str:
        return "Queries(" + self.available_queries.__repr__() + ")"

    def add_query(self, query_name: str, fn: Callable) -> None:
        """Adds a new dynamic method to this class.

        **Parameters:**

        - **query_name** - The method name as found in the SQL content.
        - **fn** - The loaded query function.
        """
        if hasattr(self, query_name):
            # this is filtered out because it can lead to hard to find bugs.
            raise SQLLoadException(f"cannot override existing attribute with query: {query_name}")
        setattr(self, query_name, fn)
        self._available_queries.add(query_name)

    def add_queries(self, queries: List[QueryFn]) -> None:
        """Add query methods to `Queries` instance."""
        for fn in queries:
            query_name = fn.__name__.rpartition(".")[2]
            self.add_query(query_name, MethodType(fn, self))

    def add_child_queries(self, child_name: str, child_queries: "Queries") -> None:
        """Adds a Queries object as a property.

        **Parameters:**

        - **child_name** - The property name to group the child queries under.
        - **child_queries** - Queries instance to add as sub-queries.
        """
        if hasattr(self, child_name):  # pragma: no cover
            # this is filtered out because it can lead to hard to find bugs.
            raise SQLLoadException(f"cannot override existing attribute with child: {child_name}")
        setattr(self, child_name, child_queries)
        for child_query_name in child_queries.available_queries:
            self._available_queries.add(f"{child_name}.{child_query_name}")

    def load_from_list(self, query_data: List[QueryDatum]):
        """Load Queries from a list of `QueryDatum`"""
        for query_datum in query_data:
            self.add_queries(self._create_methods(query_datum, self.is_aio))
        return self

    def load_from_tree(self, query_data_tree: QueryDataTree):
        """Load Queries from a `QueryDataTree`"""
        for key, value in query_data_tree.items():
            if isinstance(value, dict):
                self.add_child_queries(key, Queries(self.driver_adapter).load_from_tree(value))
            else:
                self.add_queries(self._create_methods(value, self.is_aio))
        return self



import aiosql
import sqlite3

queries = aiosql.from_path("greetings.sql", "sqlite3")

with sqlite3.connect("greetings.db") as conn:
    user = queries.get_user_by_username(conn, username="willvaughn")
    # user: (1, "willvaughn", "William")

    for _, greeting in queries.get_all_greetings(conn):
        # scan: (1, "Hi"), (2, "Aloha"), (3, "Hola"), …
        print(f"{greeting}, {user[2]}!")
    # Hi, William!
    # Aloha, William!
    # …



import pytest
from aiosql.utils import VAR_REF
from aiosql.query_loader import _UNCOMMENT, _remove_ml_comments

pytestmark = [
    pytest.mark.misc,
]


def test_var_pattern_is_quote_aware():
    sql = r"""
          select foo_id,
                 bar_id,
                 to_char(created_at, 'YYYY-MM-DD"T"HH24:MI:SSOF')
            from foos
            join bars using(bar_id)
            join bazs using(baz_id)
           where created_at < :created_at_mark
             and foo_mark > :foo_mark
        order by created_at desc, source_name asc;
    """
    groupdicts = [m.groupdict() for m in VAR_REF.finditer(sql)]
    assert len(groupdicts) == 3

    expected = [
        {
            "dquote": None,
            "lead": None,
            "squote": "'YYYY-MM-DD\"T\"HH24:MI:SSOF'",
            "var_name": None,
        },
        {
            "dquote": None,
            "lead": " ",
            "squote": None,
            "var_name": "created_at_mark",
        },
        {"dquote": None, "lead": " ", "squote": None, "var_name": "foo_mark"},
    ]
    assert groupdicts == expected


def test_var_pattern_does_not_require_semicolon_trail():
    """Make sure keywords ending queries are recognized even without
    semi-colons.
    """
    sql = r"""
        select a,
               b,
               c
          FROM foo
         WHERE a = :a"""

    groupdicts = [m.groupdict() for m in VAR_REF.finditer(sql)]
    assert len(groupdicts) == 1

    expected = {"dquote": None, "lead": " ", "squote": None, "var_name": "a"}
    assert groupdicts[0] == expected


def test_var_pattern_handles_empty_sql_string_literals():
    """Make sure SQL '' are treated correctly and don't cause a substitution to be skipped."""
    sql = r"""
        select blah
          from foo
         where lower(regexp_replace(blah,'\\W','','g')) = lower(regexp_replace(:blah,'\\W','','g'));"""

    groupdicts = [m.groupdict() for m in VAR_REF.finditer(sql)]

    expected_single_quote_match = {
        "dquote": None,
        "lead": None,
        "squote": "''",
        "var_name": None,
    }
    assert groupdicts[1] == expected_single_quote_match

    expected_var_match = {
        "dquote": None,
        "lead": "(",
        "squote": None,
        "var_name": "blah",
    }
    assert groupdicts[3] == expected_var_match


# must remove *only* OK comments
COMMENTED = """
KO
-- KO
/* OK */
'/* KO */'
"/* KO */"
' /* KO
   */'
" /* KO
   */"
/*
 * OK
 */
-- /* KO
-- KO */
/* OK
  -- OK
  ' OK ' "OK "
 */
KO
/* OK */ -- KO 'KO'
-- KO */
/*+ KO (hints must be kept!) */
"""


def test_comments():
    n = 0
    for ma in _UNCOMMENT.finditer(COMMENTED):
        matches = ma.groupdict()
        s, d, c, m = matches["squote"], matches["dquote"], matches["oneline"], matches["multiline"]
        # assert s or d or c or m, f"bad match: {m} {matches}"
        if s or d or c or m:
            n += 1
            if m:
                assert "OK" in m and "KO" not in m
            if s:
                assert "KO" in s and "OK" not in s
            if d:
                assert "KO" in d and "OK" not in d
            if c:
                assert "KO" in c and "OK" not in c
    assert n == 13


COMMENT_UNCOMMENT = [
    ("", ""),
    ("hello", "hello"),
    ("world!\n", "world!\n"),
    ("/**/", ""),
    ("/*+ hint */", "/*+ hint */"),
    ("x/*\n*/y\n", "xy\n"),
    ("-- /* */\n", "-- /* */\n"),
    ("-- /* */", "-- /* */"),
    ("'/* */'", "'/* */'"),
    ("--\n/* */X\n", "--\nX\n"),
]


def test_uncomment():
    n = 0
    for c, u in COMMENT_UNCOMMENT:
        n += 1
        assert _remove_ml_comments(c) == u
    assert n == len(COMMENT_UNCOMMENT)



import pytest
from conf_schema import USERS_DATA_PATH, BLOGS_DATA_PATH, create_user_blogs, drop_user_blogs


# guess psycopg version from a connection
def is_psycopg2(conn):
    return hasattr(conn, "get_dsn_parameters")


try:
    from pytest_postgresql import factories as pg_factories

    @pytest.fixture
    def pg_conn(request):
        """Loads seed data and return a database connection."""
        is_detached = request.config.getoption("postgresql_detached")
        if is_detached:  # pragma: no cover
            # this is *NOT* a connection, it does not have a "cursor"
            pg = request.getfixturevalue("postgresql_noproc")
            import psycopg

            conn = psycopg.connect(
                host=pg.host,
                port=pg.port,
                user=pg.user,
                password=pg.password,
                dbname=pg.dbname,
                options=pg.options,
            )
        else:
            # returns the underlying pytest-postgresql connection
            # which may be psycopg version 2 or 3, depending.
            conn = request.getfixturevalue("postgresql")

        # Loads data from blogdb fixture data
        with conn.cursor() as cur:
            for tc in create_user_blogs("pgsql"):
                cur.execute(tc)

            # guess whether we have a psycopg 2 or 3 connection
            with USERS_DATA_PATH.open() as fp:
                if is_psycopg2(conn):  # pragma: no cover
                    cur.copy_from(
                        fp, "users", sep=",", columns=["username", "firstname", "lastname"]
                    )
                else:
                    with cur.copy(
                        "COPY users(username, firstname, lastname) FROM STDIN (FORMAT CSV)"
                    ) as cope:
                        cope.write(fp.read())

            with BLOGS_DATA_PATH.open() as fp:
                if is_psycopg2(conn):  # pragma: no cover
                    cur.copy_from(
                        fp, "blogs", sep=",", columns=["userid", "title", "content", "published"]
                    )
                else:  # assume psycopg 3
                    with cur.copy(
                        "COPY blogs(userid, title, content, published) FROM STDIN (FORMAT CSV)"
                    ) as cope:
                        cope.write(fp.read())

        conn.commit()
        yield conn
        # cleanup
        with conn.cursor() as cur:
            for q in drop_user_blogs("pgsql"):
                cur.execute(q)
        conn.commit()

    @pytest.fixture
    def pg_params(request, pg_conn):
        """Build postgres connection parameters as a dictionary."""
        if is_psycopg2(pg_conn):  # pragma: no cover
            dsn = pg_conn.get_dsn_parameters()
            del dsn["tty"]
        else:  # assume psycopg 3.x
            dsn = pg_conn.info.get_parameters()
        # non empty password?
        if "password" not in dsn:
            dsn["password"] = request.config.getoption("postgresql_password") or ""
        if "port" not in dsn:
            dsn["port"] = 5432
        return dsn

    @pytest.fixture
    def pg_dsn(request, pg_params):
        """Build a postgres URL connection string."""
        p = pg_params
        yield f"postgres://{p['user']}:{p['password']}@{p['host']}:{p['port']}/{p['dbname']}"

except ModuleNotFoundError:
    # FIXME empty fixtures to please pytest

    @pytest.fixture
    def pg_conn():
        raise Exception("undefined fixture")

    @pytest.fixture
    def pg_params():
        raise Exception("undefined fixture")

    @pytest.fixture
    def pg_dsn():
        raise Exception("undefined fixture")



import argparse
import sqlite3
from datetime import datetime
from pathlib import Path
from typing import NamedTuple

import aiosql


dir_path = Path(__file__).parent
sql_path = dir_path / "sql"
db_path = dir_path / "exampleblog.db"
queries = aiosql.from_path(dir_path / "sql", "sqlite3")


users = [("bobsmith", "Bob", "Smith"), ("johndoe", "John", "Doe"), ("janedoe", "Jane", "Doe")]
blogs = [
    (
        1,
        "What I did Today",
        """\
I mowed the lawn, washed some clothes, and ate a burger.

Until next time,
Bob""",
        "2017-07-28",
    ),
    (
        3,
        "Testing",
        """\
Is this thing on?
""",
        "2018-01-01",
    ),
    (
        1,
        "How to make a pie.",
        """\
1. Make crust
2. Fill
3. Bake
4. Eat
""",
        "2018-11-23",
    ),
]


def get_conn() -> sqlite3.Connection:
    conn = sqlite3.connect(db_path)
    conn.row_factory = sqlite3.Row
    return conn


def createdb():
    if db_path.exists():
        raise SystemExit(f"Database at {db_path} already exists.")
    conn = get_conn()
    print("Inserting users and blogs data.")
    with conn:
        queries.create_schema(conn)
        queries.users.insert_many(conn, users)
        queries.blogs.insert_many(conn, blogs)
    print("Done!")
    conn.close()


def deletedb():
    print(f"Deleting the {db_path} file")
    if db_path.exists():
        db_path.unlink()


def get_users():
    conn = get_conn()
    for user in queries.users.get_all(conn):
        s = "{"
        for k in user.keys():
            s += f"{k}: {user[k]}, "
        s = s[:-2]
        s += "}"
        print(s)


def get_user_blogs(username):
    conn = get_conn()
    user_blogs = queries.blogs.get_user_blogs(conn, username=username)
    for user_blog in user_blogs:
        print("------")
        print(f"{user_blog['title']}")
        print(f"by {user_blog['author']} at {user_blog['published']}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    createdb_parser = subparsers.add_parser("createdb")
    createdb_parser.set_defaults(cmd=createdb)

    deletedb_parser = subparsers.add_parser("deletedb")
    deletedb_parser.set_defaults(cmd=deletedb)

    get_users_parser = subparsers.add_parser("get-users")
    get_users_parser.set_defaults(cmd=get_users)

    get_user_blogs_parser = subparsers.add_parser("get-user-blogs")
    get_user_blogs_parser.add_argument("username")
    get_user_blogs_parser.set_defaults(cmd=get_user_blogs)

    args = parser.parse_args()
    cmd_kwargs = {k: v for k, v in vars(args).items() if k != "cmd"}
    args.cmd(**cmd_kwargs)



import asyncio
import aiosql
import aiosqlite

queries = aiosql.from_path("greetings.sql", "aiosqlite")

async def main():
    # Parallel queries!!!
    async with aiosqlite.connect("greetings.db") as conn:
        greetings, user = await asyncio.gather(
            queries.get_all_greetings(conn),
            queries.get_user_by_username(conn, username="willvaughn")
        )

        for _, greeting in greetings:
            print(f"{greeting}, {user[2]}!")

asyncio.run(main())



from datetime import date
import aiosql
import pytest
import run_tests as t
import utils as u

try:
    import pg8000 as db
except ModuleNotFoundError:
    pytest.skip("missing driver: pg8000", allow_module_level=True)

pytestmark = [
    pytest.mark.postgres,
    pytest.mark.skipif(not u.has_pkg("pytest_postgresql"), reason="no pytest_postgresql"),
]

DRIVER = "pg8000"


@pytest.fixture
def queries():
    return t.queries(DRIVER)


@pytest.fixture
def conn(pg_params):
    dbname = pg_params["dbname"]
    del pg_params["dbname"]
    pg_params["database"] = dbname
    # cleanup unsupported keyword arguments
    for kw in ("sslcertmode", "connect_timeout", "hostaddr"):
        if kw in pg_params:
            del pg_params[kw]
    u.log.debug(f"params: {pg_params}")
    with db.connect(**pg_params) as conn:
        yield conn


def test_cursor(conn, queries):
    t.run_cursor(conn, queries)


@pytest.mark.skip("row factory is not supported")
def test_record_query(pg_params, queries):
    with db.connect(**pg_params, row_factory=dict_row) as conn:
        t.run_record_query(conn, queries)


def test_parameterized_query(conn, queries):
    t.run_parameterized_query(conn, queries)


@pytest.mark.skip("row factory is not supported")
def test_parameterized_record_query(pg_params, queries):
    with db.connect(**pg_params, row_factory=dict_row) as conn:
        t.run_parameterized_record_query(conn, queries, "pg", date)


def test_record_class_query(conn, queries):
    t.run_record_class_query(conn, queries, date)


def test_select_cursor_context_manager(conn, queries):
    t.run_select_cursor_context_manager(conn, queries, date)


def test_select_one(conn, queries):
    t.run_select_one(conn, queries)


def test_select_value(conn, queries):
    t.run_select_value(conn, queries, DRIVER)


def test_modulo(conn, queries):
    actual = queries.blogs.pg_get_modulo_2(conn, numerator=7, denominator=3)
    expected = 7 % 3
    assert actual == expected


def test_insert_returning(conn, queries):
    t.run_insert_returning(conn, queries, DRIVER, date)


def test_delete(conn, queries):
    t.run_delete(conn, queries)


def test_insert_many(conn, queries):
    t.run_insert_many(conn, queries, date)


def test_date_time(conn, queries):
    t.run_date_time(conn, queries, DRIVER)


# @pytest.mark.skip("not supported")
def test_execute_script(conn, queries):
    actual = queries.comments.pg_create_comments_table(conn)
    assert actual == "DONE"



from pathlib import Path
from typing import Callable, Dict, Optional, Type, Union, Tuple

from .adapters.aiosqlite import AioSQLiteAdapter
from .adapters.asyncpg import AsyncPGAdapter
from .adapters.pyformat import PyFormatAdapter
from .adapters.mysql import BrokenMySQLAdapter
from .adapters.generic import GenericAdapter
from .adapters.sqlite3 import SQLite3Adapter
from .adapters.pg8000 import Pg8000Adapter
from .adapters.duckdb import DuckDBAdapter
from .utils import SQLLoadException, log
from .queries import Queries
from .query_loader import QueryLoader
from .types import DriverAdapterProtocol

_ADAPTERS: Dict[str, Callable[..., DriverAdapterProtocol]] = {
    "aiosqlite": AioSQLiteAdapter,  # type: ignore
    "apsw": GenericAdapter,
    "asyncpg": AsyncPGAdapter,  # type: ignore
    "duckdb": DuckDBAdapter,
    "mariadb": BrokenMySQLAdapter,
    "mysqldb": BrokenMySQLAdapter,
    "mysql-connector": PyFormatAdapter,
    "pg8000": Pg8000Adapter,
    "psycopg": PyFormatAdapter,
    "psycopg2": PyFormatAdapter,
    "pygresql": PyFormatAdapter,
    "pymysql": BrokenMySQLAdapter,
    "sqlite3": SQLite3Adapter,
}
"""Map adapter names to their adapter class."""


def register_adapter(name: str, adapter: Callable[..., DriverAdapterProtocol]):
    """Register or override an adapter."""
    if name.lower() in _ADAPTERS:
        log.warning(f"overriding aiosql adapter {name}")
    _ADAPTERS[name.lower()] = adapter


def _make_driver_adapter(
    driver_adapter: Union[str, Callable[..., DriverAdapterProtocol]]
) -> DriverAdapterProtocol:
    """Get the driver adapter instance registered by the `driver_name`."""
    if isinstance(driver_adapter, str):
        try:
            driver_adapter = _ADAPTERS[driver_adapter.lower()]
        except KeyError:
            raise ValueError(f"Encountered unregistered driver_adapter: {driver_adapter}")
    # try some guessing if it is a PEP249 module
    elif hasattr(driver_adapter, "paramstyle"):
        style = getattr(driver_adapter, "paramstyle")  # avoid mypy warning?
        if style == "pyformat":
            driver_adapter = PyFormatAdapter  # type: ignore
        elif style == "named":
            driver_adapter = GenericAdapter  # type: ignore
        else:
            raise ValueError(f"Unexpected driver_adapter: {driver_adapter} ({style})")
    # so, can we just call it?
    if not callable(driver_adapter):
        raise ValueError(f"Unexpected driver_adapter: {driver_adapter}")

    return driver_adapter()


def from_str(
    sql: str,
    driver_adapter: Union[str, Callable[..., DriverAdapterProtocol]],
    record_classes: Optional[Dict] = None,
    kwargs_only: bool = False,
    attribute: Optional[str] = "__",
    *,
    loader_cls: Type[QueryLoader] = QueryLoader,
    queries_cls: Type[Queries] = Queries,
):
    """Load queries from a SQL string.

    **Parameters:**

    - **sql** - A string containing SQL statements and aiosql name.
    - **driver_adapter** - Either a string to designate one of the aiosql built-in database driver
      adapters. One of many available for SQLite, Postgres and MySQL. If you have defined your
      own adapter class, you can pass it's constructor.
    - **kwargs_only** - Whether to only use named parameters on query execution.
    - **attribute** - ``.`` attribute access substitution, defaults to ``"__"``, *None* disables
      the feature.
    - **record_classes** - *(optional)* **DEPRECATED** Mapping of strings used in "record_class"
      declarations to the python classes which aiosql should use when marshaling SQL results.
    - **loader_cls** - *(optional)* Custom constructor for QueryLoader extensions.
    - **queries_cls** - *(optional)* Custom constructor for Queries extensions.

    **Returns:** ``Queries``

    Usage:

    Loading queries from a SQL string.

    .. code-block:: python

      import sqlite3
      import aiosql

      sql_text = \"""
      -- name: get-all-greetings
      -- Get all the greetings in the database
      select * from greetings;

      -- name: get-user-by-username^
      -- Get all the users from the database,
      -- and return it as a dict
      select * from users where username = :username;
      \"""

      queries = aiosql.from_str(sql_text, "sqlite3")
      queries.get_all_greetings(conn)
      queries.get_user_by_username(conn, username="willvaughn")

    """
    adapter = _make_driver_adapter(driver_adapter)
    query_loader = loader_cls(adapter, record_classes, attribute=attribute)
    query_data = query_loader.load_query_data_from_sql(sql, [])
    return queries_cls(adapter, kwargs_only=kwargs_only).load_from_list(query_data)


def from_path(
    sql_path: Union[str, Path],
    driver_adapter: Union[str, Callable[..., DriverAdapterProtocol]],
    record_classes: Optional[Dict] = None,
    kwargs_only: bool = False,
    attribute: Optional[str] = "__",
    *,
    loader_cls: Type[QueryLoader] = QueryLoader,
    queries_cls: Type[Queries] = Queries,
    ext: Tuple[str] = (".sql",),
    encoding=None,
):
    """Load queries from a `.sql` file, or directory of `.sql` files.

    **Parameters:**

    - **sql_path** - Path to a `.sql` file or directory containing `.sql` files.
    - **driver_adapter** - Either a string to designate one of the aiosql built-in database driver
      adapters. One of many available for SQLite, Postgres and MySQL. If you have defined your own
      adapter class, you may pass its constructor.
    - **kwargs_only** - Whether to only use named parameters on query execution.
    - **attribute** - ``.`` attribute access substitution, defaults to ``"__""``, *None* disables
      the feature.
    - **record_classes** - *(optional)* **DEPRECATED** Mapping of strings used in "record_class"
      declarations to the python classes which aiosql should use when marshaling SQL results.
    - **loader_cls** - *(optional)* Custom constructor for `QueryLoader` extensions.
    - **queries_cls** - *(optional)* Custom constructor for `Queries` extensions.
    - **ext** - *(optional)* allowed file extensions for query files, default is `(".sql",)`.
    - **encoding** - *(optional)* encoding for reading files.

    **Returns:** `Queries`

    Usage:

    .. code-block:: python

      queries = aiosql.from_path("./sql", "psycopg2")
      queries = aiosql.from_path("./sql", MyDBAdapter)
    """
    path = Path(sql_path)

    if not path.exists():
        raise SQLLoadException(f"File does not exist: {path}")

    adapter = _make_driver_adapter(driver_adapter)
    query_loader = loader_cls(adapter, record_classes, attribute=attribute)

    if path.is_file():
        query_data = query_loader.load_query_data_from_file(path, encoding=encoding)
        return queries_cls(adapter, kwargs_only=kwargs_only).load_from_list(query_data)
    elif path.is_dir():
        query_data_tree = query_loader.load_query_data_from_dir_path(
            path, ext=ext, encoding=encoding
        )
        return queries_cls(adapter, kwargs_only=kwargs_only).load_from_tree(query_data_tree)
    else:  # pragma: no cover
        raise SQLLoadException(f"The sql_path must be a directory or file, got {sql_path}")



import asyncio
import aiosql
import aiosqlite
from typing import List

queries = aiosql.from_path("greetings.sql", "aiosqlite")

async def access_cursor():
    async with aiosqlite.connect("greetings.db") as conn:
        # append _cursor after query name
        async with queries.get_all_greetings_cursor(conn) as cur:
            print([col_info[0] for col_info in cur.description])
            first_row = await cur.fetchone()
            all_data = await cur.fetchall()
            print(f"FIRST ROW: {first_row}")  # tuple of first row data
            print(f"OTHER DATA: {all_data}")    # remaining rows

asyncio.run(access_cursor())

# ['greeting_id', 'greeting']
# FIRST ROW: (1, 'Hi')
# OTHER DATA: [(2, 'Aloha'), (3, 'Hola'), (4, 'Bonjour'), (5, '你好')]



from .aiosql import from_path, from_str, register_adapter
from .utils import SQLParseException, SQLLoadException
from importlib.metadata import version

__version__ = version("aiosql")

__all__ = ["from_path", "from_str", "register_adapter", "SQLParseException", "SQLLoadException"]



import aiosql
import psycopg2
from psycopg2.extras import execute_values

SQL = """
-- name: create_schema#
create table if not exists test (id int primary key, v1 int, v2 int);

-- name: insert!
INSERT INTO test (id, v1, v2) VALUES %s;

-- name: update!
UPDATE test SET v1 = data.v1 FROM (VALUES %s) AS data (id, v1)
WHERE test.id = data.id;

-- name: getem
select * from test order by id;
"""

queries = aiosql.from_str(SQL, "psycopg2")
conn = psycopg2.connect("dbname=test")
queries.create_schema(conn)
with conn.cursor() as cur:
    execute_values(cur, queries.insert.sql, [(1, 2, 3), (4, 5, 6), (7, 8, 9)])
    execute_values(cur, queries.update.sql, [(1, 20), (4, 50)])

print(list(queries.getem(conn)))
#  [(1, 20, 3), (4, 50, 6), (7, 8, 9)]
